from collections import defaultdict

from lato.message import Event, Task
from lato.utils import OrderedSet


class ApplicationModule:
    def __init__(self, name: str):
        self.name: str = name
        self._handlers: dict[str, set[callable]] = defaultdict(OrderedSet)
        self._submodules: OrderedSet[ApplicationModule] = OrderedSet()

    def include_submodule(self, a_module):
        assert isinstance(
            a_module, ApplicationModule
        ), f"Can only include {ApplicationModule} instances, got {a_module}"
        self._submodules.add(a_module)

    def handler(self, alias):
        """
        Decorator for registering use cases by name
        """
        try:
            is_task_or_event = issubclass(alias, (Task, Event))
        except TypeError:
            is_task_or_event = False

        if callable(alias) and not is_task_or_event:
            # @app.handle(my_function)
            func = alias
            alias = func.__name__
            assert len(self._handlers[alias]) == 0
            self._handlers[alias].add(func)
            return func

        # @app.handle("my_function")
        # @app.handle(MyTask)
        def decorator(func):
            """
            Decorator for registering use cases by name
            """
            assert len(self._handlers[alias]) == 0
            self._handlers[alias].add(func)
            return func

        return decorator

    def iterate_handlers_for(self, alias: str):
        if alias in self._handlers:
            for handler in self._handlers[alias]:
                yield handler
        for submodule in self._submodules:
            try:
                yield from submodule.iterate_handlers_for(alias)
            except KeyError:
                pass

    def on(self, event_name):
        # TODO: add matcher parameter
        def decorator(func):
            """
            Decorator for registering an event handler

            :param event_handler:
            :return:
            """
            self._handlers[event_name].add(func)
            return func

        return decorator

    def __repr__(self):
        return f"<{self.name} {object.__repr__(self)}>"



from typing import Any, Callable

from lato.application_module import ApplicationModule
from lato.dependency_provider import SimpleDependencyProvider
from lato.message import Event, Task
from lato.transaction_context import TransactionContext


class Application(ApplicationModule):
    dependency_provider_class = SimpleDependencyProvider

    def __init__(self, name=__name__, dependency_provider=None, **kwargs):
        super().__init__(name)
        self.dependency_provider = (
            dependency_provider or self.dependency_provider_class(**kwargs)
        )
        self._transaction_context_factory = None
        self._on_enter_transaction_context = lambda ctx: None
        self._on_exit_transaction_context = lambda ctx, exception=None: None
        self._transaction_middlewares = []

    def get_dependency(self, identifier: Any) -> Any:
        """Get a dependency from the dependency provider"""
        return self.dependency_provider.get_dependency(identifier)

    def __getitem__(self, item) -> Any:
        return self.get_dependency(item)

    def call(self, func: Callable | str, *args, **kwargs):
        with self.transaction_context() as ctx:
            result = ctx.call(func, *args, **kwargs)
        return result

    def execute(self, task: Task) -> Any:
        with self.transaction_context() as ctx:
            result = ctx.execute(task)
        return result

    def emit(self, event: Event) -> dict[callable, Any]:
        with self.transaction_context() as ctx:
            result = ctx.emit(event)
        return result

    def on_enter_transaction_context(self, func):
        """
        Decorator for registering a function to be called when entering a transaction context

        :param func:
        :return:
        """
        self._on_enter_transaction_context = func
        return func

    def on_exit_transaction_context(self, func):
        """
        Decorator for registering a function to be called when exiting a transaction context

        :param func:
        :return:
        """
        self._on_exit_transaction_context = func
        return func

    def on_create_transaction_context(self, func):
        """
        Decorator for overrinding default transaction context creation

        :param func:
        :return:
        """
        self._transaction_context_factory = func
        return func

    def transaction_middleware(self, middleware_func):
        """
        Decorator for registering a middleware function to be called when executing a function in a transaction context
        :param middleware_func:
        :return:
        """
        self._transaction_middlewares.insert(0, middleware_func)
        return middleware_func

    def transaction_context(self, **dependencies) -> TransactionContext:
        """
        Creates a transaction context with the application dependencies

        :param dependencies:
        :return:
        """
        if self._transaction_context_factory:
            ctx = self._transaction_context_factory(**dependencies)
        else:
            dp = self.dependency_provider.copy(**dependencies)
            ctx = TransactionContext(dependency_provider=dp)

        ctx.configure(
            on_enter_transaction_context=self._on_enter_transaction_context,
            on_exit_transaction_context=self._on_exit_transaction_context,
            middlewares=self._transaction_middlewares,
            handlers_iterator=self.iterate_handlers_for,
        )
        return ctx



from uuid import UUID, uuid4

from pydantic import BaseModel, Field


class Task(BaseModel):
    task_id: UUID = Field(default_factory=uuid4, alias="task_id")


class Event(BaseModel):
    event_id: UUID = Field(default_factory=uuid4, alias="event_id")
